import json
import logging
import subprocess
import sys
from datetime import datetime, timezone
from getpass import getpass
from pathlib import Path
from typing import TYPE_CHECKING

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

from fastapi.testclient import TestClient

from yak_server.config_file import get_settings
from yak_server.database import Base, SessionLocal, engine, mysql_settings
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    MatchReferenceModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)

if TYPE_CHECKING:
    from fastapi import FastAPI

logger = logging.getLogger(__name__)


class ConfirmPasswordDoesNotMatch(Exception):
    def __init__(self) -> None:
        super().__init__("Password and Confirm Password fields does not match.")


class SignupError(Exception):
    def __init__(self, description: str) -> None:
        super().__init__(f"Error during signup. {description}")


class RecordDeletionInProduction(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to delete records in production using script. DO NOT DO IT.")


class TableDropInProduction(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to drop database tables in production using script. DO NOT DO IT.")


class BackupError(Exception):
    def __init__(self, description: str) -> None:
        super().__init__(f"Error during backup. {description}")


def create_database() -> None:
    Base.metadata.create_all(bind=engine)


def create_admin(app: "FastAPI") -> None:
    password = getpass(prompt="Admin user password: ")
    confirm_password = getpass(prompt="Confirm admin password: ")

    if password != confirm_password:
        raise ConfirmPasswordDoesNotMatch

    client = TestClient(app)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": "admin",
            "first_name": "admin",
            "last_name": "admin",
            "password": password,
        },
    )

    if not response_signup.json()["ok"]:
        raise SignupError(response_signup.json()["description"])


def initialize_database(app: "FastAPI") -> None:
    db = SessionLocal()

    data_folder = get_settings().data_folder

    with Path(f"{data_folder}/phases.json").open() as file:
        phases = json.loads(file.read())

        db.add_all(PhaseModel(**phase) for phase in phases)
        db.flush()

    with Path(f"{data_folder}/groups.json").open() as file:
        groups = json.loads(file.read())

        for group in groups:
            phase = db.query(PhaseModel).filter_by(code=group["phase_code"]).first()
            group.pop("phase_code")
            group["phase_id"] = phase.id

        db.add_all(GroupModel(**group) for group in groups)
        db.flush()

    with Path(f"{data_folder}/teams.json").open() as file:
        teams = json.loads(file.read())

        for team in teams:
            team["flag_url"] = ""

            team_instance = TeamModel(**team)
            db.add(team_instance)
            db.flush()

            team_instance.flag_url = app.url_path_for(
                "retrieve_team_flag_by_id",
                team_id=team_instance.id,
            )
            db.flush()

    with Path(f"{data_folder}/matches.json").open() as file:
        matches = json.loads(file.read())

        for match in matches:
            if match["team1_code"] is None:
                match["team1_id"] = None
            else:
                team1 = db.query(TeamModel).filter_by(code=match["team1_code"]).first()
                match["team1_id"] = team1.id

            if match["team2_code"] is None:
                match["team2_id"] = None
            else:
                team2 = db.query(TeamModel).filter_by(code=match["team2_code"]).first()
                match["team2_id"] = team2.id

            group = db.query(GroupModel).filter_by(code=match["group_code"]).first()
            match["group_id"] = group.id

            match.pop("team1_code")
            match.pop("team2_code")
            match.pop("group_code")

        db.add_all(MatchReferenceModel(**match) for match in matches)
        db.flush()

    db.commit()


def backup_database() -> None:
    with resources.as_file(resources.files("yak_server") / "cli/backup_files") as path:
        backup_location = path

    Path(backup_location).mkdir(exist_ok=True)

    backup_datetime = datetime.now(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%f%z")

    file_name = f"{backup_location}/yak_toto_backup_{backup_datetime}.sql"

    result = subprocess.run(
        (
            f"mysqldump {mysql_settings.db} "
            f"-u {mysql_settings.user_name} "
            f"-P {mysql_settings.port} "
            "--protocol=tcp "
            f"--password='{mysql_settings.password}'"
        ),
        shell=True,
        capture_output=True,
        encoding="utf-8",
    )

    if result.returncode:
        error_message = (
            f"Something went wrong when backup on {backup_datetime}: "
            f"{result.stderr.replace(mysql_settings.password, '********')}"
        )

        logger.error(error_message)

        raise BackupError(error_message)

    with Path(file_name).open(mode="w") as file:
        file.write(result.stdout)
        logger.info(f"Backup done on {backup_datetime}")


def delete_database(app: "FastAPI") -> None:
    if not app.debug:
        raise RecordDeletionInProduction

    db = SessionLocal()

    db.query(GroupPositionModel).delete()
    db.query(ScoreBetModel).delete()
    db.query(BinaryBetModel).delete()
    db.query(UserModel).delete()
    db.query(MatchReferenceModel).delete()
    db.query(MatchModel).delete()
    db.query(GroupModel).delete()
    db.query(PhaseModel).delete()
    db.query(TeamModel).delete()
    db.commit()


def drop_database(app: "FastAPI") -> None:
    if not app.debug:
        raise TableDropInProduction

    Base.metadata.drop_all(bind=engine)


def setup_migration() -> None:
    with resources.as_file(resources.files("yak_server") / ".." / "alembic.ini") as path:
        print(
            "To be able to run the database migration "
            "scripts, you need to run the following command:",
        )
        print(f"export ALEMBIC_CONFIG='{path.resolve()}'")
        print()
        print(
            "Follow this link for more informations: "
            "https://alembic.sqlalchemy.org/en/latest/tutorial.html#editing-the-ini-file",
        )



from sqlalchemy.orm import Session

from yak_server.database import SessionLocal


def get_db() -> Session:
    db_connection = SessionLocal()

    try:
        yield db_connection
    finally:
        db_connection.close()



import json
import secrets
import sys
from configparser import ConfigParser
from getpass import getpass
from pathlib import Path
from uuid import UUID

from dateutil import parser

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

from yak_server.config_file import RuleContainer, Rules
from yak_server.database import MySQLSettings
from yak_server.helpers.rules import RULE_MAPPING


class RuleNotDefined(Exception):
    def __init__(self, rule_id: str) -> None:
        super().__init__(f"Rule not defined: {rule_id}")


class EnvBuilder:
    def __init__(self) -> None:
        self.env = {}
        self.env_mysql = {}

        debug = input("DEBUG (y/n): ")

        self.debug = debug == "y"
        self.env["DEBUG"] = 1 if self.debug else 0

    def setup_profiling(self) -> None:
        if self.debug:
            profiling = input("PROFILING (y/n): ")

            self.env["PROFILING"] = 1 if profiling == "y" else 0

    def setup_mysql_env(self) -> None:
        user_name = input("MYSQL USER NAME: ")
        password = getpass("MYSQL PASSWORD: ")
        port = input("MYSQL PORT (default: 3306): ")
        db = input("MYSQL DB: ")

        # try to instance pydantic model to check if
        # settings are ok
        if not port:
            MySQLSettings(user_name=user_name, password=password, db=db)
        else:
            MySQLSettings(user_name=user_name, password=password, port=port, db=db)

        self.env_mysql["MYSQL_USER_NAME"] = user_name
        self.env_mysql["MYSQL_PASSWORD"] = password
        if port:
            self.env_mysql["MYSQL_PORT"] = port
        self.env_mysql["MYSQL_DB"] = db

    def setup_jwt(self) -> None:
        jwt_expiration_time = input("JWT_EXPIRATION_TIME: ")

        self.env["JWT_EXPIRATION_TIME"] = jwt_expiration_time
        self.env["JWT_SECRET_KEY"] = secrets.token_hex(128)

    def choose_competition(self) -> None:
        # Select competition to load associated rules
        with resources.as_file(resources.files("yak_server") / "data") as path:
            available_competitions = []

            for competition in path.glob("*"):
                available_competitions.append(competition.stem)

            for index, competition in enumerate(available_competitions, 1):
                print(f"{index} - {competition}")

            competition_choice = int(input("Choose your competition: "))

            competition = available_competitions[competition_choice - 1]

            data_folder = path / competition

        # Load rules in environment
        rules = Rules(__root__=[])

        for rule_file in Path(f"{data_folder}/rules").glob("*.json"):
            rule_id = UUID(rule_file.stem)

            if rule_id not in RULE_MAPPING:
                raise RuleNotDefined(rule_id)

            with rule_file.open() as rule_content:
                rules.__root__.append(
                    RuleContainer(id=rule_id, config=json.loads(rule_content.read())),
                )

        # Load configuration to compute points
        config = ConfigParser()
        config.read(f"{data_folder}/config.ini")

        self.env["COMPETITION"] = competition
        self.env["LOCK_DATETIME"] = parser.parse(config.get("locking", "datetime"))
        self.env["BASE_CORRECT_RESULT"] = config.getint("points", "base_correct_result")
        self.env["MULTIPLYING_FACTOR_CORRECT_RESULT"] = config.getint(
            "points",
            "multiplying_factor_correct_result",
        )
        self.env["BASE_CORRECT_SCORE"] = config.getint("points", "base_correct_score")
        self.env["MULTIPLYING_FACTOR_CORRECT_SCORE"] = config.getint(
            "points",
            "multiplying_factor_correct_score",
        )
        self.env["TEAM_QUALIFIED"] = config.getint("points", "team_qualified")
        self.env["FIRST_TEAM_QUALIFIED"] = config.getint("points", "first_team_qualified")
        self.env["DATA_FOLDER"] = data_folder
        self.env["RULES"] = rules.json()

    def write(self) -> None:
        write_env_file(self.env, ".env")
        write_env_file(self.env_mysql, ".env.mysql")


def write_env_file(env: dict, filename: str) -> None:
    with Path(f"./{filename}").open(mode="w") as file:
        for env_var, env_value in env.items():
            file.write(f"{env_var}={env_value}\n")


def init_env() -> None:
    env_builder = EnvBuilder()
    env_builder.setup_profiling()
    env_builder.setup_mysql_env()
    env_builder.setup_jwt()
    env_builder.choose_competition()

    env_builder.write()



from typing import TYPE_CHECKING, List, Optional

from pydantic import UUID4, BaseModel

from yak_server.helpers.language import Lang, get_language_description

from .generic import PositiveOrZeroInt

if TYPE_CHECKING:
    from yak_server.database.models import TeamModel


class FlagOut(BaseModel):
    url: str


class TeamIn(BaseModel):
    id: UUID4
    score: Optional[PositiveOrZeroInt]


class TeamOut(BaseModel):
    id: UUID4
    code: str
    description: str
    flag: FlagOut

    @classmethod
    def from_instance(cls, team: "TeamModel", lang: Lang) -> "TeamOut":
        return cls(
            id=team.id,
            code=team.code,
            description=get_language_description(team, lang),
            flag=FlagOut(url=team.flag_url),
        )


class AllTeamsResponse(BaseModel):
    teams: List[TeamOut]


class OneTeamResponse(BaseModel):
    team: TeamOut


class TeamWithWonOut(BaseModel):
    id: UUID4
    code: str
    description: str
    flag: FlagOut
    won: Optional[bool]


class TeamWithScoreOut(BaseModel):
    id: UUID4
    code: str
    description: str
    flag: FlagOut
    score: Optional[PositiveOrZeroInt]


class TeamModifyScoreBetIn(BaseModel):
    score: Optional[PositiveOrZeroInt]


class TeamModifyBinaryBetIn(BaseModel):
    id: Optional[UUID4]



from typing import Optional

from fastapi import Depends
from sqlalchemy.orm import Session
from strawberry.fastapi import BaseContext

from yak_server.config_file import Settings, get_settings
from yak_server.database.models import UserModel
from yak_server.v1.helpers.database import get_db


class YakContext(BaseContext):
    db: Session
    settings: Settings
    user: Optional[UserModel]

    def __init__(self, db: Session, settings: Settings, user: Optional[UserModel] = None) -> None:
        super().__init__()
        self.db = db
        self.settings = settings
        self.user = user


def get_context(
    db: Session = Depends(get_db),
    settings: Settings = Depends(get_settings),
) -> YakContext:
    return YakContext(db, settings)



from typing import TYPE_CHECKING

from pydantic import BaseModel, PositiveInt

from .generic import PositiveOrZeroInt

if TYPE_CHECKING:
    from yak_server.database.models import UserModel


class UserResult(BaseModel):
    rank: PositiveInt
    first_name: str
    last_name: str
    full_name: str
    number_match_guess: PositiveOrZeroInt
    number_score_guess: PositiveOrZeroInt
    number_qualified_teams_guess: PositiveOrZeroInt
    number_first_qualified_guess: PositiveOrZeroInt
    number_quarter_final_guess: PositiveOrZeroInt
    number_semi_final_guess: PositiveOrZeroInt
    number_final_guess: PositiveOrZeroInt
    number_winner_guess: PositiveOrZeroInt
    points: float

    @classmethod
    def from_instance(cls, user: "UserModel", rank: PositiveInt) -> "UserResult":
        return cls(
            rank=rank,
            first_name=user.first_name,
            last_name=user.last_name,
            full_name=user.full_name,
            number_match_guess=user.number_match_guess,
            number_score_guess=user.number_score_guess,
            number_qualified_teams_guess=user.number_qualified_teams_guess,
            number_first_qualified_guess=user.number_first_qualified_guess,
            number_quarter_final_guess=user.number_quarter_final_guess,
            number_semi_final_guess=user.number_semi_final_guess,
            number_final_guess=user.number_final_guess,
            number_winner_guess=user.number_winner_guess,
            points=user.points,
        )



"""Remove played column. Compute played property as won + drawn + lost.

Revision ID: d75e76959af8
Revises: 9f8e020eeced
Create Date: 2023-03-12 01:50:45.579007

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "d75e76959af8"
down_revision = "9f8e020eeced"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("group_position", schema=None) as batch_op:
        batch_op.drop_column("played")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("group_position", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("played", mysql.INTEGER(), autoincrement=False, nullable=False),
        )

    # ### end Alembic commands ###



from typing import List
from uuid import UUID

import strawberry

from yak_server.helpers.errors import (
    EXPIRED_TOKEN_MESSAGE,
    INVALID_CREDENTIALS_MESSAGE,
    INVALID_TOKEN_MESSAGE,
    LOCKED_BINARY_BET_MESSAGE,
    LOCKED_SCORE_BET_MESSAGE,
    UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
    binary_bet_not_found_message,
    group_not_found_message,
    name_already_exists_message,
    phase_not_found_message,
    score_bet_not_found_message,
    team_not_found_message,
    user_not_found_message,
)

from .schema import (
    BinaryBet,
    Group,
    GroupPosition,
    Phase,
    ScoreBet,
    Team,
    User,
    UserWithoutSensitiveInfo,
    UserWithToken,
)


@strawberry.type
class InvalidToken:
    message: str = INVALID_TOKEN_MESSAGE


@strawberry.type
class ExpiredToken:
    message: str = EXPIRED_TOKEN_MESSAGE


@strawberry.type
class UnauthorizedAccessToAdminAPI:
    message: str = UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE


@strawberry.type
class LockedScoreBetError:
    message: str = LOCKED_SCORE_BET_MESSAGE


@strawberry.type
class ScoreBetNotFoundForUpdate:
    message: str = "Score bet not found. Cannot modify a resource that does not exist."


@strawberry.type
class NewScoreNegative:
    variable_name: strawberry.Private[str]
    score: strawberry.Private[int]

    @strawberry.field
    def message(self) -> str:
        return (
            f"Variable '{self.variable_name}' got invalid value {self.score}. "
            "Score cannot be negative."
        )


ModifyScoreBetResult = strawberry.union(
    "ModifyScoreBetResult",
    types=(
        ScoreBet,
        ScoreBetNotFoundForUpdate,
        LockedScoreBetError,
        NewScoreNegative,
        InvalidToken,
        ExpiredToken,
    ),
)


@strawberry.type
class LockedBinaryBetError:
    message: str = LOCKED_BINARY_BET_MESSAGE


@strawberry.type
class BinaryBetNotFoundForUpdate:
    message: str = "Binary bet not found. Cannot modify a resource that does not exist."


ModifyBinaryBetResult = strawberry.union(
    "ModifyBinaryBetResult",
    types=(
        BinaryBet,
        BinaryBetNotFoundForUpdate,
        LockedBinaryBetError,
        InvalidToken,
        ExpiredToken,
    ),
)


CurrentUserResult = strawberry.union("CurrentUserResult", types=(User, InvalidToken, ExpiredToken))


@strawberry.type
class AllTeamsSuccessful:
    teams: List[Team]


AllTeamsResult = strawberry.union(
    "AllTeamsResult",
    types=(AllTeamsSuccessful, InvalidToken, ExpiredToken),
)


@strawberry.type
class TeamByIdNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return team_not_found_message(self.id)


@strawberry.type
class TeamByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return team_not_found_message(self.code)


TeamByIdResult = strawberry.union(
    "TeamByIdResult",
    types=(Team, TeamByIdNotFound, InvalidToken, ExpiredToken),
)
TeamByCodeResult = strawberry.union(
    "TeamByCodeResult",
    types=(Team, TeamByCodeNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class ScoreBetNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return score_bet_not_found_message(self.id)


ScoreBetResult = strawberry.union(
    "ScoreBetResult",
    types=(ScoreBet, ScoreBetNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class BinaryBetNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return binary_bet_not_found_message(self.id)


BinaryBetResult = strawberry.union(
    "BinaryBetResult",
    types=(
        BinaryBet,
        BinaryBetNotFound,
        InvalidToken,
        ExpiredToken,
    ),
)


@strawberry.type
class Groups:
    groups: List[Group]


AllGroupsResult = strawberry.union("AllGroupsResult", types=(Groups, InvalidToken, ExpiredToken))


@strawberry.type
class GroupByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return group_not_found_message(self.code)


GroupByCodeResult = strawberry.union(
    "GroupByCodeResult",
    types=(Group, GroupByCodeNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class GroupByIdNotFound:
    id: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return group_not_found_message(self.id)


GroupByIdResult = strawberry.union(
    "GroupByIdResult",
    types=(Group, GroupByIdNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class Phases:
    phases: List[Phase]


AllPhasesResult = strawberry.union("AllPhasesResult", types=(Phases, InvalidToken, ExpiredToken))


@strawberry.type
class PhaseByCodeNotFound:
    code: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return phase_not_found_message(self.code)


PhaseByCodeResult = strawberry.union(
    "PhaseByCodeResult",
    types=(Phase, PhaseByCodeNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class PhaseByIdNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return phase_not_found_message(self.id)


PhaseByIdResult = strawberry.union(
    "PhaseByIdResult",
    types=(Phase, PhaseByIdNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class ScoreBoard:
    users: List[UserWithoutSensitiveInfo]


ScoreBoardResult = strawberry.union(
    "ScoreBoardResult",
    types=(ScoreBoard, InvalidToken, ExpiredToken),
)


@strawberry.type
class UserNameAlreadyExists:
    user_name: strawberry.Private[str]

    @strawberry.field
    def message(self) -> str:
        return name_already_exists_message(self.user_name)


SignupResult = strawberry.union("SignupResult", types=(UserWithToken, UserNameAlreadyExists))


@strawberry.type
class InvalidCredentials:
    message: str = INVALID_CREDENTIALS_MESSAGE


LoginResult = strawberry.union("LoginResult", types=(UserWithToken, InvalidCredentials))


@strawberry.type
class GroupRank:
    group_rank: List[GroupPosition]
    group: Group


GroupRankByCodeResult = strawberry.union(
    "GroupRankByCodeResult",
    types=(GroupRank, GroupByCodeNotFound, InvalidToken, ExpiredToken),
)


GroupRankByIdResult = strawberry.union(
    "GroupRankByIdResult",
    types=(GroupRank, GroupByIdNotFound, InvalidToken, ExpiredToken),
)


@strawberry.type
class UserNotFound:
    id: strawberry.Private[UUID]

    @strawberry.field
    def message(self) -> str:
        return user_not_found_message(self.id)


ModifyUserResult = strawberry.union(
    "ModifyUserResult",
    types=(
        UserWithoutSensitiveInfo,
        UserNotFound,
        InvalidToken,
        ExpiredToken,
        UnauthorizedAccessToAdminAPI,
    ),
)



from typing import TYPE_CHECKING, Optional

from pydantic import UUID4, BaseModel, PositiveInt

from yak_server.helpers.language import Lang, get_language_description

from .groups import GroupIn, GroupOut
from .phases import PhaseOut
from .teams import FlagOut, TeamIn, TeamModifyBinaryBetIn, TeamWithWonOut

if TYPE_CHECKING:
    from yak_server.database.models import BinaryBetModel


class BinaryBetIn(BaseModel):
    is_one_won: Optional[bool]
    index: PositiveInt
    team1: TeamIn
    team2: TeamIn
    group: GroupIn


class BinaryBetOut(BaseModel):
    id: UUID4
    locked: bool
    team1: Optional[TeamWithWonOut]
    team2: Optional[TeamWithWonOut]

    @classmethod
    def from_instance(
        cls,
        binary_bet: "BinaryBetModel",
        locked: bool,
        lang: "Lang",
    ) -> "BinaryBetOut":
        return cls(
            id=binary_bet.id,
            locked=locked,
            team1=TeamWithWonOut(
                id=binary_bet.match.team1.id,
                code=binary_bet.match.team1.code,
                description=get_language_description(binary_bet.match.team1, lang),
                won=binary_bet.bet_from_is_one_won()[0],
                flag=FlagOut(url=binary_bet.match.team1.flag_url),
            )
            if binary_bet.match.team1
            else None,
            team2=TeamWithWonOut(
                id=binary_bet.match.team2.id,
                code=binary_bet.match.team2.code,
                description=get_language_description(binary_bet.match.team2, lang),
                won=binary_bet.bet_from_is_one_won()[1],
                flag=FlagOut(url=binary_bet.match.team2.flag_url),
            )
            if binary_bet.match.team2
            else None,
        )


class Group(BaseModel):
    id: UUID4


class BinaryBetWithGroupIdOut(BaseModel):
    id: UUID4
    locked: bool
    group: Group
    team1: Optional[TeamWithWonOut]
    team2: Optional[TeamWithWonOut]

    @classmethod
    def from_instance(
        cls,
        binary_bet: "BinaryBetModel",
        locked: bool,
        lang: Lang,
    ) -> "BinaryBetWithGroupIdOut":
        return cls(
            id=binary_bet.id,
            locked=locked,
            group=Group(id=binary_bet.match.group_id),
            team1=TeamWithWonOut(
                id=binary_bet.match.team1.id,
                code=binary_bet.match.team1.code,
                description=get_language_description(binary_bet.match.team1, lang),
                won=binary_bet.bet_from_is_one_won()[0],
                flag=FlagOut(url=binary_bet.match.team1.flag_url),
            )
            if binary_bet.match.team1
            else None,
            team2=TeamWithWonOut(
                id=binary_bet.match.team2.id,
                code=binary_bet.match.team2.code,
                description=get_language_description(binary_bet.match.team2, lang),
                won=binary_bet.bet_from_is_one_won()[1],
                flag=FlagOut(url=binary_bet.match.team2.flag_url),
            )
            if binary_bet.match.team2
            else None,
        )


class BinaryBetResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    binary_bet: BinaryBetOut


class ModifyBinaryBetIn(BaseModel):
    is_one_won: Optional[bool]
    team1: Optional[TeamModifyBinaryBetIn]
    team2: Optional[TeamModifyBinaryBetIn]

    class Config:
        extra = "forbid"



from datetime import datetime
from functools import lru_cache
from typing import List
from uuid import UUID

from pydantic import BaseModel, BaseSettings


class RuleContainer(BaseModel):
    id: UUID
    config: dict


class Rules(BaseModel):
    __root__: List[RuleContainer]


class Settings(BaseSettings):
    jwt_secret_key: str
    jwt_expiration_time: int
    competition: str
    lock_datetime: datetime
    base_correct_result: int
    multiplying_factor_correct_result: int
    base_correct_score: int
    multiplying_factor_correct_score: int
    team_qualified: int
    first_team_qualified: int
    data_folder: str
    rules: Rules

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"


@lru_cache(maxsize=None)
def get_settings() -> Settings:
    return Settings()



import click

from yak_server import create_app

from .database import (
    backup_database,
    create_admin,
    create_database,
    delete_database,
    drop_database,
    initialize_database,
    setup_migration,
)
from .env import init_env


@click.command()
def create() -> None:
    """Create all database tables."""
    create_database()


@click.command()
def init() -> None:
    """Initialize database."""
    app = create_app()
    initialize_database(app)


@click.command()
def drop() -> None:
    """Drop all tables."""
    app = create_app()
    drop_database(app)


@click.command()
def delete() -> None:
    """Delete all records."""
    app = create_app()
    delete_database(app)


@click.command()
def admin() -> None:
    """Create admin account in database."""
    app = create_app()
    create_admin(app)


@click.command()
def backup() -> None:
    """Backup database in a sql file."""
    backup_database()


@click.command()
def migration() -> None:
    """Help to run database migration scripts"""
    setup_migration()


@click.group()
def db() -> None:
    pass


db.add_command(create)
db.add_command(init)
db.add_command(drop)
db.add_command(delete)
db.add_command(admin)
db.add_command(backup)
db.add_command(migration)


@click.command()
def init() -> None:
    """Build the env files you need to start the server"""
    init_env()


@click.group()
def env() -> None:
    pass


env.add_command(init)


@click.group()
def main() -> None:
    pass


main.add_command(db)
main.add_command(env)



import re


def is_uuid4(uuid: str) -> bool:
    regex = re.compile(
        r"^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}\Z",
    )
    match = regex.match(uuid)
    return bool(match)


def is_iso_3166_1_alpha_2_code(code: str) -> bool:
    regex = re.compile(r"^([A-Z]{2}|[A-Z]{2}-[A-Z]{3})\Z")
    match = regex.match(code)
    return bool(match)



import logging

from fastapi import APIRouter, Depends
from pydantic import UUID4
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from yak_server.config_file import Settings, get_settings
from yak_server.database.models import (
    BinaryBetModel,
    MatchModel,
    UserModel,
)
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging import modify_binary_bet_successfully
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.database import get_db
from yak_server.v1.helpers.errors import (
    BetNotFound,
    GroupNotFound,
    LockedBinaryBet,
    TeamNotFound,
)
from yak_server.v1.models.binary_bets import (
    BinaryBetIn,
    BinaryBetOut,
    BinaryBetResponse,
    ModifyBinaryBetIn,
    TeamWithWonOut,
)
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.teams import FlagOut

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/binary_bets",
    tags=["binary_bets"],
)


def send_response(
    binary_bet: BinaryBetModel,
    locked: bool,
    lang: Lang,
) -> GenericOut[BinaryBetResponse]:
    return GenericOut(
        result=BinaryBetResponse(
            phase=PhaseOut.from_instance(binary_bet.match.group.phase, lang),
            group=GroupOut.from_instance(binary_bet.match.group, lang),
            binary_bet=BinaryBetOut(
                id=binary_bet.id,
                locked=locked,
                team1=TeamWithWonOut(
                    id=binary_bet.match.team1.id,
                    code=binary_bet.match.team1.code,
                    description=get_language_description(binary_bet.match.team1, lang),
                    flag=FlagOut(url=binary_bet.match.team1.flag_url),
                    won=binary_bet.bet_from_is_one_won()[0],
                )
                if binary_bet.match.team1
                else None,
                team2=TeamWithWonOut(
                    id=binary_bet.match.team2.id,
                    code=binary_bet.match.team2.code,
                    description=get_language_description(binary_bet.match.team2, lang),
                    flag=FlagOut(url=binary_bet.match.team2.flag_url),
                    won=binary_bet.bet_from_is_one_won()[1],
                )
                if binary_bet.match.team2
                else None,
            ),
        ),
    )


@router.post("/")
def create_binary_bet(
    binary_bet_in: BinaryBetIn,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[BinaryBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedBinaryBet

    match = MatchModel(
        team1_id=binary_bet_in.team1.id,
        team2_id=binary_bet_in.team2.id,
        index=binary_bet_in.index,
        group_id=binary_bet_in.group.id,
    )

    db.add(match)
    try:
        db.flush()
    except IntegrityError as integrity_error:
        db.rollback()
        if "FOREIGN KEY (`team1_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=binary_bet_in.team1.id) from integrity_error

        if "FOREIGN KEY (`team2_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=binary_bet_in.team2.id) from integrity_error

        raise GroupNotFound(group_id=binary_bet_in.group.id) from integrity_error

    binary_bet = BinaryBetModel(
        match_id=match.id,
        user_id=user.id,
        is_one_won=binary_bet_in.is_one_won,
    )

    db.add(binary_bet)
    db.commit()
    db.refresh(binary_bet)

    return send_response(binary_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.get("/{bet_id}")
def retrieve_binary_bet_by_id(
    bet_id: UUID4,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[BinaryBetResponse]:
    binary_bet = db.query(BinaryBetModel).filter_by(user_id=user.id, id=str(bet_id)).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    return send_response(binary_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.patch("/{bet_id}")
def modify_binary_bet_by_id(
    bet_id: UUID4,
    modify_binary_bet_in: ModifyBinaryBetIn,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[BinaryBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedBinaryBet

    binary_bet = db.query(BinaryBetModel).filter_by(user_id=user.id, id=str(bet_id)).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    logger.info(
        modify_binary_bet_successfully(user.name, binary_bet, modify_binary_bet_in.is_one_won),
    )

    if "is_one_won" in modify_binary_bet_in.__fields_set__:
        binary_bet.is_one_won = modify_binary_bet_in.is_one_won

    if modify_binary_bet_in.team1 is not None and "id" in modify_binary_bet_in.team1.__fields_set__:
        binary_bet.match.team1_id = modify_binary_bet_in.team1.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            raise TeamNotFound(modify_binary_bet_in.team1.id) from integrity_error

    if modify_binary_bet_in.team2 is not None and "id" in modify_binary_bet_in.team2.__fields_set__:
        binary_bet.match.team2_id = modify_binary_bet_in.team2.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            raise TeamNotFound(modify_binary_bet_in.team2.id) from integrity_error

    db.commit()

    return send_response(binary_bet, is_locked(user.name, settings.lock_datetime), lang)


@router.delete("/{bet_id}")
def delete_binary_bet_by_id(
    bet_id: UUID4,
    lang: Lang = DEFAULT_LANGUAGE,
    db: Session = Depends(get_db),
    user: UserModel = Depends(get_current_user),
    settings: Settings = Depends(get_settings),
) -> GenericOut[BinaryBetResponse]:
    if is_locked(user.name, settings.lock_datetime):
        raise LockedBinaryBet

    binary_bet = db.query(BinaryBetModel).filter_by(user_id=user.id, id=str(bet_id)).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    response = send_response(binary_bet, is_locked(user.name, settings.lock_datetime), lang)

    db.delete(binary_bet)
    db.commit()

    return response



from functools import wraps
from typing import TYPE_CHECKING

from jwt import ExpiredSignatureError, PyJWTError

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import decode_bearer_token

from .result import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI

NUMBER_ELEMENTS_IN_AUTHORIZATION = 2

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from yak_server.config_file import Settings


def is_authenticated(f):  # noqa: ANN001, ANN201
    @wraps(f)
    def _verify(*args, **kwargs):
        auth_headers = kwargs["info"].context.request.headers.get("Authorization", "").split()
        db: "Session" = kwargs["info"].context.db
        settings: "Settings" = kwargs["info"].context.settings

        if len(auth_headers) != NUMBER_ELEMENTS_IN_AUTHORIZATION or auth_headers[0] != "Bearer":
            return InvalidToken()

        token = auth_headers[1]
        try:
            data = decode_bearer_token(token, settings.jwt_secret_key)
        except ExpiredSignatureError:
            return ExpiredToken()
        except PyJWTError:
            return InvalidToken()

        user = db.query(UserModel).filter_by(id=data["sub"]).first()
        if not user:
            return InvalidToken()

        kwargs["info"].context.user = user

        return f(*args, **kwargs)

    return _verify


def is_admin_authenticated(f):  # noqa: ANN001, ANN201
    @wraps(f)
    def _verify(*args, **kwargs):
        user: UserModel = kwargs["info"].context.user

        if user.name != "admin":
            return UnauthorizedAccessToAdminAPI()

        return f(*args, **kwargs)

    return _verify



from datetime import datetime, timedelta, timezone
from typing import Any, Dict

from jwt import decode as jwt_decode
from jwt import encode as jwt_encode


def encode_bearer_token(sub: str, expiration_time: timedelta, secret_key: str) -> str:
    return jwt_encode(
        {
            "sub": sub,
            "iat": datetime.now(tz=timezone.utc),
            "exp": datetime.now(tz=timezone.utc) + expiration_time,
        },
        secret_key,
        algorithm="HS512",
    )


def decode_bearer_token(token: str, secret_key: str) -> Dict[str, Any]:
    return jwt_decode(token, secret_key, algorithms=["HS512"])



from typing import TYPE_CHECKING, List
from uuid import UUID

import strawberry
from strawberry.types import Info

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.group_position import compute_group_rank

from .bearer_authentication import is_authenticated
from .result import (
    AllGroupsResult,
    AllPhasesResult,
    AllTeamsResult,
    AllTeamsSuccessful,
    BinaryBetNotFound,
    BinaryBetResult,
    CurrentUserResult,
    GroupByCodeNotFound,
    GroupByCodeResult,
    GroupByIdNotFound,
    GroupByIdResult,
    GroupRank,
    GroupRankByCodeResult,
    GroupRankByIdResult,
    Groups,
    PhaseByCodeNotFound,
    PhaseByCodeResult,
    PhaseByIdNotFound,
    PhaseByIdResult,
    Phases,
    ScoreBetNotFound,
    ScoreBetResult,
    ScoreBoard,
    ScoreBoardResult,
    TeamByCodeNotFound,
    TeamByCodeResult,
    TeamByIdNotFound,
    TeamByIdResult,
    User,
)
from .schema import (
    BinaryBet,
    Group,
    Phase,
    ScoreBet,
    Team,
    UserWithoutSensitiveInfo,
    send_group_position,
)

if TYPE_CHECKING:
    from datetime import datetime

    from sqlalchemy.orm import Session

    from yak_server.config_file import Settings


@strawberry.type
class Query:
    @strawberry.field
    @is_authenticated
    def current_user_result(self, info: Info) -> CurrentUserResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        return User.from_instance(db=db, instance=user, lock_datetime=settings.lock_datetime)

    @strawberry.field
    @is_authenticated
    def all_teams_result(self, info: Info) -> AllTeamsResult:
        db: "Session" = info.context.db

        return AllTeamsSuccessful(
            teams=[Team.from_instance(instance=team) for team in db.query(TeamModel).all()],
        )

    @strawberry.field
    @is_authenticated
    def team_by_id_result(self, id: UUID, info: Info) -> TeamByIdResult:
        db: "Session" = info.context.db

        team_record = db.query(TeamModel).filter_by(id=str(id)).first()

        if not team_record:
            return TeamByIdNotFound(id=id)

        return Team.from_instance(instance=team_record)

    @strawberry.field
    @is_authenticated
    def team_by_code_result(self, code: str, info: Info) -> TeamByCodeResult:
        db: "Session" = info.context.db

        team_record = db.query(TeamModel).filter_by(code=code).first()

        if not team_record:
            return TeamByCodeNotFound(code=code)

        return Team.from_instance(instance=team_record)

    @strawberry.field
    @is_authenticated
    def score_bet_result(self, id: UUID, info: Info) -> ScoreBetResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        score_bet_record = db.query(ScoreBetModel).filter_by(id=str(id), user_id=user.id).first()

        if not score_bet_record:
            return ScoreBetNotFound(id=id)

        return ScoreBet.from_instance(
            db=db,
            instance=score_bet_record,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def binary_bet_result(self, id: UUID, info: Info) -> BinaryBetResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        binary_bet_record = db.query(BinaryBetModel).filter_by(id=str(id), user_id=user.id).first()

        if not binary_bet_record:
            return BinaryBetNotFound(id=id)

        return BinaryBet.from_instance(
            db=db,
            instance=binary_bet_record,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def all_groups_result(self, info: Info) -> AllGroupsResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        return Groups(
            groups=[
                Group.from_instance(
                    db=db,
                    instance=group,
                    user_id=user.id,
                    lock_datetime=settings.lock_datetime,
                )
                for group in db.query(GroupModel).order_by(GroupModel.index)
            ],
        )

    @strawberry.field
    @is_authenticated
    def group_by_id_result(self, id: UUID, info: Info) -> GroupByIdResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        group_record = db.query(GroupModel).filter_by(id=str(id)).first()

        if not group_record:
            return GroupByIdNotFound(id=id)

        return Group.from_instance(
            db=db,
            instance=group_record,
            user_id=user.id,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def group_by_code_result(self, code: strawberry.ID, info: Info) -> GroupByCodeResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        group_record = db.query(GroupModel).filter_by(code=code).first()

        if not group_record:
            return GroupByCodeNotFound(code=code)

        return Group.from_instance(
            db=db,
            instance=group_record,
            user_id=user.id,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def all_phases_result(self, info: Info) -> AllPhasesResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        phases = db.query(PhaseModel).order_by(PhaseModel.index)

        return Phases(
            phases=[
                Phase.from_instance(
                    db=db,
                    instance=phase,
                    user_id=user.id,
                    lock_datetime=settings.lock_datetime,
                )
                for phase in phases
            ],
        )

    @strawberry.field
    @is_authenticated
    def phase_by_id_result(self, id: UUID, info: Info) -> PhaseByIdResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        phase_record = db.query(PhaseModel).filter_by(id=str(id)).first()

        if not phase_record:
            return PhaseByIdNotFound(id=id)

        return Phase.from_instance(
            db=db,
            instance=phase_record,
            user_id=user.id,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def phase_by_code_result(self, code: str, info: Info) -> PhaseByCodeResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        phase_record = db.query(PhaseModel).filter_by(code=code).first()

        if not phase_record:
            return PhaseByCodeNotFound(code=code)

        return Phase.from_instance(
            db=db,
            instance=phase_record,
            user_id=user.id,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    @is_authenticated
    def score_board_result(self, info: Info) -> ScoreBoardResult:
        db: "Session" = info.context.db

        users = db.query(UserModel).filter(UserModel.name != "admin")

        return ScoreBoard(
            users=[UserWithoutSensitiveInfo.from_instance(instance=user) for user in users],
        )

    @strawberry.field
    @is_authenticated
    def group_rank_by_code_result(self, code: str, info: Info) -> GroupRankByCodeResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        group = db.query(GroupModel).filter_by(code=code).first()

        if not group:
            return GroupByCodeNotFound(code=code)

        group_rank = db.query(GroupPositionModel).filter_by(
            user_id=user.id,
            group_id=group.id,
        )

        def send_response(
            db: "Session",
            user_id: UUID,
            group: GroupModel,
            group_rank: List[GroupPositionModel],
            lock_datetime: "datetime",
        ) -> GroupRank:
            return GroupRank(
                group_rank=send_group_position(group_rank),
                group=Group.from_instance(
                    db=db,
                    instance=group,
                    user_id=user_id,
                    lock_datetime=lock_datetime,
                ),
            )

        if not any(group_position.need_recomputation for group_position in group_rank):
            return send_response(db, user.id, group, group_rank, settings.lock_datetime)

        score_bets = user.score_bets.filter(MatchModel.group_id == group.id).join(
            ScoreBetModel.match,
        )
        group_rank = compute_group_rank(group_rank, score_bets)
        db.commit()

        return send_response(db, user.id, group, group_rank, settings.lock_datetime)

    @strawberry.field
    @is_authenticated
    def group_rank_by_id_result(self, id: UUID, info: Info) -> GroupRankByIdResult:
        db: "Session" = info.context.db
        user: UserModel = info.context.user
        settings: "Settings" = info.context.settings

        group = db.query(GroupModel).filter_by(id=str(id)).first()

        if not group:
            return GroupByIdNotFound(id=id)

        group_rank = db.query(GroupPositionModel).filter_by(
            user_id=user.id,
            group_id=group.id,
        )

        def send_response(
            db: "Session",
            user_id: UUID,
            group: GroupModel,
            group_rank: List[GroupPositionModel],
            lock_datetime: "datetime",
        ) -> GroupRank:
            return GroupRank(
                group_rank=send_group_position(group_rank),
                group=Group.from_instance(
                    db=db,
                    instance=group,
                    user_id=user_id,
                    lock_datetime=lock_datetime,
                ),
            )

        if not any(group_position.need_recomputation for group_position in group_rank):
            return send_response(db, user.id, group, group_rank, settings.lock_datetime)

        score_bets = user.score_bets.filter(MatchModel.group_id == group.id).join(
            ScoreBetModel.match,
        )
        group_rank = compute_group_rank(group_rank, score_bets)
        db.commit()

        return send_response(db, user.id, group, group_rank, settings.lock_datetime)



from functools import partial

INVALID_CREDENTIALS_MESSAGE = "Invalid credentials"
UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE = "Unauthorized access to admin API"
INVALID_TOKEN_MESSAGE = "Invalid token, authentication required"
EXPIRED_TOKEN_MESSAGE = "Expired token, re-authentication required"
LOCKED_SCORE_BET_MESSAGE = "Cannot modify score bet, lock date is exceeded"
LOCKED_BINARY_BET_MESSAGE = "Cannot modify binary bet, lock date is exceeded"


def name_already_exists_message(user_name: str) -> str:
    return f"Name already exists: {user_name}"


def generic_not_found_message(id: str, resource_name: str) -> str:
    return f"{resource_name} not found: {id}"


bet_not_found_message = partial(generic_not_found_message, resource_name="Bet")
binary_bet_not_found_message = partial(generic_not_found_message, resource_name="Binary bet")
group_not_found_message = partial(generic_not_found_message, resource_name="Group")
phase_not_found_message = partial(generic_not_found_message, resource_name="Phase")
rule_not_found_message = partial(generic_not_found_message, resource_name="Rule")
score_bet_not_found_message = partial(generic_not_found_message, resource_name="Score bet")
team_not_found_message = partial(generic_not_found_message, resource_name="Team")
user_not_found_message = partial(generic_not_found_message, resource_name="User")
